"use strict";
/* eslint key-spacing:0 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Projectz = void 0;
// Imports
// First we need to import the libraries we require.
// Load in the file system libraries
const fs_list_1 = __importDefault(require("@bevry/fs-list"));
const fs_read_1 = __importDefault(require("@bevry/fs-read"));
const fs_write_1 = __importDefault(require("@bevry/fs-write"));
const path_1 = require("path");
// Handle configuration files
const json_1 = require("@bevry/json");
// [TypeChecker](https://github.com/bevry/typechecker) is used for checking data types
const typechecker_1 = require("typechecker");
// Fellow handling, and contributor fetching
const fellow_1 = __importDefault(require("fellow"));
// Load in our other project files
const backer_js_1 = require("./backer.js");
const badge_js_1 = require("./badge.js");
const history_js_1 = require("./history.js");
const install_js_1 = require("./install.js");
const license_js_1 = require("./license.js");
const util_js_1 = require("./util.js");
__exportStar(require("./types.js"), exports);
// Definition
// Projects is defined as a class to ensure we can run multiple instances of it
class Projectz {
    /** our log function to use (logLevel, ...messages) */
    log = function () { };
    /** the current working directory (the path) that projectz is working on */
    cwd;
    /**
     * The absolute paths for all the package files.
     * Should be arranged in the order of merging preference.
     */
    filenamesForPackageFiles = {
        component: null,
        bower: null,
        jquery: null,
        package: null,
        projectz: null,
    };
    /** the data for each of our package files */
    dataForPackageFiles = {};
    /** the absolute paths for all the meta files */
    filenamesForReadmeFiles = {
        // gets filled in with relative paths
        readme: null,
        history: null,
        contributing: null,
        backers: null,
        license: null,
    };
    /** the data for each of our readme files */
    dataForReadmeFiles = {};
    // Apply options
    constructor(opts) {
        this.cwd = opts.cwd ? (0, path_1.resolve)(opts.cwd) : process.cwd();
        if (opts.log)
            this.log = opts.log;
        // this.log = console.log.bind(console)
    }
    /** Compile the project */
    async compile() {
        // Load readme and package data
        await this.loadPaths();
        // Enhance our package data
        const enhancedPackagesData = await this.enhancePackagesData();
        // Enhance our readme data
        const enhancedReadmesData = await this.enhanceReadmesData(enhancedPackagesData);
        // Save
        await this.save(enhancedPackagesData, enhancedReadmesData);
    }
    /** Load in the paths we have specified */
    async loadPaths() {
        // Apply our determined paths for packages
        const packages = Object.keys(this.filenamesForPackageFiles);
        const ReadmeFiles = Object.keys(this.filenamesForReadmeFiles);
        // Load
        const files = await (0, fs_list_1.default)(this.cwd);
        for (const file of files) {
            const filePath = (0, path_1.join)(this.cwd, file);
            for (const key of packages) {
                const basename = file.toLowerCase().split('.').slice(0, -1).join('.');
                if (basename === key) {
                    this.log('info', `Reading package file: ${filePath}`);
                    const data = await (0, json_1.readJSON)(filePath);
                    this.filenamesForPackageFiles[key] = file;
                    this.dataForPackageFiles[key] = data;
                }
            }
            for (const key of ReadmeFiles) {
                if (file.toLowerCase().startsWith(key)) {
                    this.log('info', `Reading meta file: ${filePath}`);
                    const data = await (0, fs_read_1.default)(filePath);
                    this.filenamesForReadmeFiles[key] = file;
                    this.dataForReadmeFiles[key] = data.toString();
                }
            }
        }
    }
    /** Merge and enhance the packages data */
    async enhancePackagesData() {
        // ----------------------------------
        // Combine
        this.log('debug', 'Enhancing packages data');
        // Combine the package data
        const mergedPackagesData = {
            keywords: [],
            editions: [],
            badges: {
                list: [],
                config: {},
            },
            bugs: {},
            readmes: {},
            packages: {},
            repository: {},
            github: {},
            dependencies: {},
            devDependencies: {},
        };
        for (const key of Object.keys(this.filenamesForPackageFiles)) {
            Object.assign(mergedPackagesData, this.dataForPackageFiles[key]);
        }
        // ----------------------------------
        // Validation
        // Validate keywords field
        if ((0, typechecker_1.isString)(mergedPackagesData.keywords)) {
            throw new Error('projectz: keywords field must be array instead of CSV');
        }
        // Validate sponsors array
        if (mergedPackagesData.sponsor) {
            throw new Error('projectz: sponsor field is deprecated, use sponsors field');
        }
        if ((0, typechecker_1.isString)(mergedPackagesData.sponsors)) {
            throw new Error('projectz: sponsors field must be array instead of CSV');
        }
        // Validate maintainers array
        if (mergedPackagesData.maintainer) {
            throw new Error('projectz: maintainer field is deprecated, use maintainers field');
        }
        if ((0, typechecker_1.isString)(mergedPackagesData.maintainers)) {
            throw new Error('projectz: maintainers field must be array instead of CSV');
        }
        // Validate license SPDX string
        if ((0, typechecker_1.isPlainObject)(mergedPackagesData.license)) {
            throw new Error('projectz: license field must now be a valid SPDX string: https://docs.npmjs.com/files/package.json#license');
        }
        if ((0, typechecker_1.isPlainObject)(mergedPackagesData.licenses)) {
            throw new Error('projectz: licenses field is deprecated, you must now use the license field as a valid SPDX string: https://docs.npmjs.com/files/package.json#license');
        }
        // Validate enhanced fields
        const objs = ['badges', 'readmes', 'packages', 'github', 'bugs'];
        for (const key of objs) {
            if (!(0, typechecker_1.isPlainObject)(mergedPackagesData[key])) {
                throw new Error(`projectz: ${key} property must be an object`);
            }
        }
        // Validate package values
        for (const [key, value] of Object.entries(mergedPackagesData.packages)) {
            if (!(0, typechecker_1.isPlainObject)(value)) {
                throw new Error(`projectz: custom package data for package ${key} must be an object`);
            }
        }
        // Validate badges field
        if (!Array.isArray(mergedPackagesData.badges.list) ||
            (mergedPackagesData.badges.config &&
                !(0, typechecker_1.isPlainObject)(mergedPackagesData.badges.config))) {
            throw new Error('projectz: badges field must be in the format of: {list: [], config: {}}\nSee https://github.com/bevry/badges for details.');
        }
        mergedPackagesData.badges.config ??= {};
        // ----------------------------------
        // Ensure
        // Ensure repository is an object
        if (typeof mergedPackagesData.repository === 'string') {
            const githubSlug = (0, util_js_1.getGithubSlug)(mergedPackagesData);
            if (githubSlug) {
                mergedPackagesData.repository = {
                    type: 'git',
                    url: `https://github.com/${githubSlug}.git`,
                };
            }
        }
        // Fallback name
        if (!mergedPackagesData.name) {
            mergedPackagesData.name = (0, path_1.dirname)(this.cwd);
        }
        // Fallback version
        if (!mergedPackagesData.version) {
            mergedPackagesData.version = '0.1.0';
        }
        // Fallback demo field, by scanning homepage
        if (!mergedPackagesData.demo && mergedPackagesData.homepage) {
            mergedPackagesData.demo = mergedPackagesData.homepage;
        }
        // Fallback title from name
        if (!mergedPackagesData.title) {
            mergedPackagesData.title = mergedPackagesData.name;
        }
        // Fallback description
        if (!mergedPackagesData.description) {
            mergedPackagesData.description = 'no description was provided';
        }
        // Fallback browsers field, by checking if `component` or `bower` package files exists, or if the `browser` or `jspm` fields are defined
        if (mergedPackagesData.browsers == null) {
            mergedPackagesData.browsers = Boolean(this.filenamesForPackageFiles.bower ||
                this.filenamesForPackageFiles.component ||
                mergedPackagesData.browser ||
                mergedPackagesData.jspm);
        }
        // ----------------------------------
        // Enhance Respository
        // Converge and extract repository information
        let github;
        if (mergedPackagesData.repository) {
            const githubSlug = (0, util_js_1.getGithubSlug)(mergedPackagesData);
            if (githubSlug) {
                // Extract parts
                const [githubUsername, githubRepository] = githubSlug.split('/');
                const githubUrl = 'https://github.com/' + githubSlug;
                const githubRepositoryUrl = githubUrl + '.git';
                // Github data
                github = {
                    username: githubUsername,
                    repository: githubRepository,
                    slug: githubSlug,
                    url: githubUrl,
                    repositoryUrl: githubRepositoryUrl,
                };
                // Badges
                Object.assign(mergedPackagesData.badges.config, {
                    githubUsername,
                    githubRepository,
                    githubSlug,
                });
                // Fallback bugs field by use of repo
                if (!mergedPackagesData.bugs) {
                    mergedPackagesData.bugs = github && {
                        url: `https://github.com/${github.slug}/issues`,
                    };
                }
                // Fetch contributors
                // await getContributorsFromRepo(githubSlug)
            }
        }
        // ----------------------------------
        // Enhance People
        // Fellows
        const authors = fellow_1.default.add(mergedPackagesData.authors || mergedPackagesData.author);
        const contributors = fellow_1.default.add(mergedPackagesData.contributors).filter((fellow) => fellow.name.includes('[bot]') === false);
        const maintainers = fellow_1.default.add(mergedPackagesData.maintainers).filter((fellow) => fellow.name.includes('[bot]') === false);
        const sponsors = fellow_1.default.add(mergedPackagesData.sponsors);
        // ----------------------------------
        // Enhance Packages
        // Create the data for the `package.json` format
        const pkg = Object.assign(
        // New Object
        {}, 
        // Old Data
        this.dataForPackageFiles.package || {}, 
        // Enhanced Data
        {
            name: mergedPackagesData.name,
            version: mergedPackagesData.version,
            license: mergedPackagesData.license,
            description: mergedPackagesData.description,
            keywords: mergedPackagesData.keywords,
            author: (0, util_js_1.getPeopleTextArray)(authors, {
                displayEmail: true,
                displayYears: true,
            }).join(', '),
            maintainers: (0, util_js_1.getPeopleTextArray)(maintainers, {
                displayEmail: true,
                urlFields: ['githubUrl', 'url'],
            }),
            contributors: (0, util_js_1.getPeopleTextArray)(contributors, {
                displayEmail: true,
                urlFields: ['githubUrl', 'url'],
            }),
            repository: mergedPackagesData.repository,
            bugs: mergedPackagesData.bugs,
            engines: mergedPackagesData.engines,
            dependencies: mergedPackagesData.dependencies,
            devDependencies: mergedPackagesData.devDependencies,
            main: mergedPackagesData.main,
        }, 
        // Explicit data
        mergedPackagesData.packages.package || {});
        // Trim
        // @ts-ignore
        if ((0, typechecker_1.isEmptyPlainObject)(pkg.dependencies))
            delete pkg.dependencies;
        // @ts-ignore
        if ((0, typechecker_1.isEmptyPlainObject)(pkg.devDependencies))
            delete pkg.devDependencies;
        // Badges
        if (!mergedPackagesData.badges.config.npmPackageName && pkg.name) {
            mergedPackagesData.badges.config.npmPackageName = pkg.name;
        }
        // Create the data for the `jquery.json` format, which is essentially the same as the `package.json` format so just extend that
        const jquery = Object.assign(
        // New Object
        {}, 
        // Old Data
        this.dataForPackageFiles.jquery || {}, 
        // Enhanced Data
        pkg, 
        // Explicit data
        mergedPackagesData.packages.jquery || {});
        // Create the data for the `component.json` format
        const component = Object.assign(
        // New Object
        {}, 
        // Old Data
        this.dataForPackageFiles.component || {}, 
        // Enhanced Data
        {
            name: mergedPackagesData.name,
            version: mergedPackagesData.version,
            license: mergedPackagesData.license,
            description: mergedPackagesData.description,
            keywords: mergedPackagesData.keywords,
            demo: mergedPackagesData.demo,
            main: mergedPackagesData.main,
            scripts: [mergedPackagesData.main],
        }, 
        // Explicit data
        mergedPackagesData.packages.component || {});
        // Create the data for the `bower.json` format
        const bower = Object.assign(
        // New Object
        {}, 
        // Old Data
        this.dataForPackageFiles.bower || {}, 
        // Enhanced Data
        {
            name: mergedPackagesData.name,
            version: mergedPackagesData.version,
            license: mergedPackagesData.license,
            description: mergedPackagesData.description,
            keywords: mergedPackagesData.keywords,
            authors: (0, util_js_1.getPeopleTextArray)(authors, {
                displayYears: true,
                displayEmail: true,
            }),
            main: mergedPackagesData.main,
        }, 
        // Explicit data
        mergedPackagesData.packages.bower || {});
        // ----------------------------------
        // Enhance Combination
        // Merge together
        const enhancedPackagesData = Object.assign({}, mergedPackagesData, {
            // Add paths so that our helpers have access to them
            filenamesForPackageFiles: this.filenamesForPackageFiles,
            filenamesForReadmeFiles: this.filenamesForReadmeFiles,
            // Other
            github,
            // Fellows
            authors,
            contributors,
            maintainers,
            sponsors,
            // Create the data for the `package.json` format
            package: pkg,
            jquery,
            component,
            bower,
        });
        // Return
        return enhancedPackagesData;
    }
    /** Merge and enhance the readmes data */
    async enhanceReadmesData(data) {
        const enhancedReadmesData = {};
        /* eslint prefer-const: 0 */
        for (let [key, value] of Object.entries(this.dataForReadmeFiles)) {
            if (!value) {
                this.log('debug', `Enhancing readme value: ${key} — skipped`);
                continue;
            }
            value = (0, util_js_1.replaceSection)(['TITLE', 'NAME'], value, `<h1>${data.title}</h1>`);
            value = (0, util_js_1.replaceSection)(['BADGES', 'BADGE'], value, badge_js_1.getBadgesSection.bind(null, data));
            value = (0, util_js_1.replaceSection)(['DESCRIPTION'], value, data.description);
            value = (0, util_js_1.replaceSection)(['INSTALL'], value, install_js_1.getInstallInstructions.bind(null, data));
            value = (0, util_js_1.replaceSection)(['CONTRIBUTE', 'CONTRIBUTING'], value, data.github
                ? backer_js_1.getContributeSection.bind(null, data)
                : '<!-- github projects only -->');
            value = (0, util_js_1.replaceSection)(['BACKERS', 'BACKER'], value, data.github
                ? backer_js_1.getBackerSection.bind(null, data)
                : '<!-- github projects only -->');
            value = (0, util_js_1.replaceSection)(['BACKERSFILE', 'BACKERFILE'], value, data.github
                ? backer_js_1.getBackerFile.bind(null, data)
                : '<!-- github projects only -->');
            value = (0, util_js_1.replaceSection)(['HISTORY', 'CHANGES', 'CHANGELOG'], value, data.github
                ? history_js_1.getHistorySection.bind(null, data)
                : '<!-- github projects only -->');
            value = (0, util_js_1.replaceSection)(['LICENSE', 'LICENSES'], value, data.github
                ? license_js_1.getLicenseSection.bind(null, data)
                : '<!-- github projects only -->');
            value = (0, util_js_1.replaceSection)(['LICENSEFILE'], value, license_js_1.getLicenseFile.bind(null, data));
            enhancedReadmesData[key] = (0, util_js_1.trim)(value) + '\n';
            this.log('info', `Enhanced readme value: ${key}`);
        }
        return enhancedReadmesData;
    }
    /** Save the data we've loaded into the files */
    async save(enhancedPackagesData, enhancedReadmesData) {
        // Prepare
        this.log('info', 'Writing changes...');
        await Promise.all([
            // save package files
            ...Object.entries(this.filenamesForPackageFiles).map(async ([key, filename]) => {
                if (!filename || key === 'projectz')
                    return;
                const filepath = (0, path_1.join)(this.cwd, filename);
                this.log('info', `Saving package file: ${filepath}`);
                const data = enhancedPackagesData[key];
                return (0, json_1.writeJSON)(filepath, data);
            }),
            // save readme files
            ...Object.entries(this.filenamesForReadmeFiles).map(async ([key, filename]) => {
                if (!filename)
                    return;
                const filepath = (0, path_1.join)(this.cwd, filename);
                this.log('info', `Saving readme file: ${filepath}`);
                const content = enhancedReadmesData[key];
                return (0, fs_write_1.default)(filepath, content);
            }),
        ]);
        // log
        this.log('info', 'Wrote changes');
    }
}
exports.Projectz = Projectz;
